/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.management.AttributeChangeNotification;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.apache.commons.pool.PoolUtils;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationNotifier;

/**
 * The JBI Worker manager is in charge of creating the JBI Workers and the
 * associated ThreadPool.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class JBIWorkerManager implements WorkerPoolManager {

    private ThreadPoolExecutor workersThreadPool;

    private GenericObjectPool JBIWorkersPool;

    private AbstractComponent component;

    private RuntimeConfigurationNotifier runtimeConfiguration;

    private Logger logger;

    private NotificationListener listener = new NotificationListener() {
        /*
         * (non-Javadoc)
         * 
         * @see
         * javax.management.NotificationListener#handleNotification(javax.management
         * .Notification, java.lang.Object)
         */
        public void handleNotification(Notification notification, Object obj) {
            if (notification instanceof AttributeChangeNotification) {
                AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
                logger.log(Level.FINE, "Notification received for attribute : "
                        + attrNotif.getAttributeName());
                if (attrNotif.getAttributeName().equals("processorPoolSize")) {
                    Integer newVal = (Integer) (attrNotif.getNewValue());
                    setCorePoolSize(newVal.intValue());
                }
            }
        }
    };

    /**
     * Creates a new instance of {@link JBIWorkerManager}
     * 
     * @param component
     * @param configuration
     * @param logger
     */
    public JBIWorkerManager(final AbstractComponent component,
            final RuntimeConfigurationNotifier runtimeConfiguration, final Logger logger) {
        this.component = component;
        this.runtimeConfiguration = runtimeConfiguration;
        this.logger = logger;

        this.runtimeConfiguration.addNotificationListener(listener, null, null);
    }
    
    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.listener.PoolManager#start()
     */
    public void start() throws JBIException {
        this.createJBIListenerPool();
        this.createJBIWorkersPool();
        this.startJBIWorkersPool();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.listener.PoolManager#stop()
     */
    public void stop() throws JBIException {
        // try to let the executing task finish
        this.workersThreadPool.shutdown();
        try {
            this.workersThreadPool.awaitTermination(60, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            this.logger.log(Level.SEVERE, "Workers pool interrupted : " + e.getMessage());
        }
        // interrupt the remaining executing tasks
        this.workersThreadPool.shutdownNow();
        
        //clear the pool of JBI listeners
        this.JBIWorkersPool.clear();
    }

    /**
     * Create the workers pool.
     * 
     */
    protected void createJBIWorkersPool() {
        this.logger.fine("Creating JBI workers pool, core pool size is "
                + this.component.getComponentConfiguration().getProcessorPoolSize());
        this.workersThreadPool = new ThreadPoolExecutor(this.component.getComponentConfiguration()
                .getProcessorPoolSize().getValue(), 255, 60, TimeUnit.SECONDS,
                new LinkedBlockingQueue<Runnable>(1024), new JBIListenerThreadFactory(component
                        .getContext().getComponentName(), this.logger));
    }

    /**
	 * 
	 */
    protected void startJBIWorkersPool() {
        this.workersThreadPool.prestartAllCoreThreads();
    }

    /**
     * Creates a pool of JBIListeners instances. This is done to avoid JBI
     * workers to create a new JBIListener instance on each request since the
     * {@link AbstractJBIListener} implementation can consume many time in the
     * implemented methods.
     * 
     */
    protected void createJBIListenerPool() {
        logger.log(Level.FINE, "Creating JBI listeners object pool, size is "
                + this.component.getComponentConfiguration().getProcessorPoolSize());
        this.JBIWorkersPool = new GenericObjectPool(new JBIListenerPoolableObjectFactory(
                this.component), this.component.getComponentConfiguration().getProcessorPoolSize()
                .getValue(), GenericObjectPool.WHEN_EXHAUSTED_GROW,
                GenericObjectPool.DEFAULT_MAX_WAIT, this.component.getComponentConfiguration()
                        .getProcessorPoolSize().getValue(), GenericObjectPool.DEFAULT_MIN_IDLE,
                false, false, GenericObjectPool.DEFAULT_TIME_BETWEEN_EVICTION_RUNS_MILLIS,
                GenericObjectPool.DEFAULT_NUM_TESTS_PER_EVICTION_RUN,
                GenericObjectPool.DEFAULT_MIN_EVICTABLE_IDLE_TIME_MILLIS,
                GenericObjectPool.DEFAULT_TEST_WHILE_IDLE);

        // initialize the pool with N objects
        try {
            PoolUtils.prefill(this.JBIWorkersPool, this.component.getComponentConfiguration()
                    .getProcessorPoolSize().getValue());
        } catch (IllegalArgumentException e1) {
        } catch (Exception e1) {
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.listener.WorkersPoolManager#getThreadPool
     * ()
     */
    public ThreadPoolExecutor getThreadPool() {
        return this.workersThreadPool;
    }

    /**
     * Get the JBIWorkers Pool. Contains {@link AbstractJBIListener}
     * implementation instances.
     * 
     * @return
     */
    public GenericObjectPool getWorkersPool() {
        return this.JBIWorkersPool;
    }

    /**
     * 
     * @param size
     */
    private void setCorePoolSize(int size) {
        if (workersThreadPool != null) {
            System.out.println("Reset the processor thread pool size to " + size);
            this.logger.info("Reset the processor thread pool size to " + size);
            workersThreadPool.setCorePoolSize(size);
        }
    }
}
